"""config.py

Config defines the config for the Rebuilder and a marshmallow schema for
validating the config.

Author:
    Alex Potter-Dixon <apotter-dixon@glasswallsolutions.com>
"""
from typing import List

from marshmallow import EXCLUDE, Schema, fields, post_load
from victoria.encryption.schemas import EncryptionEnvelopeSchema, EncryptionEnvelope


def compare_encryption_envelopes(e1: EncryptionEnvelope, e2: EncryptionEnvelope) -> bool:
    return e1.data == e2.data and e1.iv == e2.iv and e1.key == e2.key and e1.version == e2.version


class AccessSchema(Schema):
    """Marshmallow schema for the Accessing AzureDevops plugin config."""

    access_token = fields.Nested(EncryptionEnvelopeSchema, required=True, allow_none=False)
    organisation = fields.Nested(EncryptionEnvelopeSchema, required=True, allow_none=False)
    project = fields.Nested(EncryptionEnvelopeSchema, required=True, allow_none=False)
    email = fields.Nested(EncryptionEnvelopeSchema, required=True, allow_none=False)

    @post_load
    def create_access_config(self, data, **kwargs):
        return EncryptedAccessConfig(**data)


class EncryptedAccessConfig:
    """AccessConfig is the config for accessing Azure Devops.

    Attributes:
        access_token (EncryptionEnvelope): The access token for the Azure DevOps API.
        organisation (EncryptionEnvelope): The Azure DevOps organisation to use.
        project (EncryptionEnvelope): The Azure DevOps plugin to use.
        email (EncryptionEnvelope): The email the user uses with Azure DevOps.
    """

    def __init__(self, access_token: EncryptionEnvelope, organisation: EncryptionEnvelope, project: EncryptionEnvelope,
                 email: EncryptionEnvelope) -> None:
        self.access_token = access_token
        self.organisation = organisation
        self.project = project
        self.email = email

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return compare_encryption_envelopes(self.access_token, other.access_token) \
                   and compare_encryption_envelopes(self.organisation, other.organisation) \
                   and compare_encryption_envelopes(self.project, other.project) \
                   and compare_encryption_envelopes(self.email, other.email)
        return False


class AccessConfig:
    """AccessConfig is the config for accessing Azure Devops.

    Attributes:
        access_token (str): The access token for the Azure DevOps API.
        organisation (str): The Azure DevOps organisation to use.
        project (str): The Azure DevOps plugin to use.
        email (str): The email the user uses with Azure DevOps.
    """

    def __init__(self, access_token: str, organisation: str, project: str,
                 email: str) -> None:
        self.access_token = access_token
        self.organisation = organisation
        self.project = project
        self.email = email

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.access_token == other.access_token \
                   and self.organisation == other.organisation \
                   and self.project == other.project \
                   and self.email == other.email
        return False


class ReleaseSchema(Schema):
    """Schema for releases"""
    name = fields.Str()

    @post_load
    def create_release_config(self, data, **kwargs):
        return ReleaseConfig(**data)


class ReleaseConfig:
    """ReleaseConfig is the config for releases in Azure Devops.

    Attributes:
        name (str): Name of the release.

    Parameters:
        complete (bool): If a release is complete.
        release_id (int): The release id of the release.
        environment_id (int): The id of the environment associated with the release.
    """

    def __init__(self, name: str) -> None:
        self.name = name
        self.complete = False
        self.release_id = 0
        self.environment_id = 0

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.name == other.name
        return False


class DeploymentSchema(Schema):
    """Marshmallow schema for deployments."""
    stage = fields.Str()
    releases = fields.List(fields.Nested(ReleaseSchema))

    @post_load
    def create_deployment_config(self, data, **kwargs):
        return DeploymentConfig(**data)


class DeploymentConfig:
    """StageConfig is the config for stages.

    Attributes:
        releases (List[ReleaseConfig]): The list of releases to deploy.
    """

    def __init__(self, releases: List[ReleaseConfig], stage: str) -> None:
        self.releases = releases
        self.stage = stage
        self.complete = False

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.releases == other.releases and self.stage == other.stage
        return False


class RebuilderSchema(Schema):
    """Marshmallow schema for rebuilder."""
    access = fields.Nested(AccessSchema)
    deployments = fields.List(fields.Nested(DeploymentSchema))
    environments = fields.List(fields.Str)

    @post_load
    def create_rebuilder_config(self, data, **kwargs):
        """Callback used by marshmallow after loading object. We're using it here
        to create an instance of Config after loading the data."""
        return RebuilderConfig(**data)


class RebuilderConfig:
    """DeploymentConfig is the config for deployments.

    Attributes:
        stages (List[StageConfig}): List of stage configurations.
    """

    def __init__(
            self,
            access: EncryptedAccessConfig,
            deployments: List[DeploymentConfig],
    ) -> None:
        self.access = access
        self.deployments = deployments

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.access == other.access \
                   and self.deployments == other.deployments
        return False


CONFIG_SCHEMA = RebuilderSchema(unknown=EXCLUDE)
"""Instance of ConfigSchema to use for validation."""



"""
A further wrapper on the AzureDevops Python API to abstract common interactions
with Azure Devops such as checking if a release is complete.

Parameters:
    access_cfg (AccessConfig): The configuration to access AzureDevOps.

"""

import logging
from typing import Tuple, Union

from azure.devops.connection import Connection
from msrest.authentication import BasicAuthentication

from victoria_rebuilder.config import AccessConfig


class DevOpsClient:
    def __init__(self, access_cfg: AccessConfig):

        self.access_cfg = access_cfg
        self._connect()

    def get_release_status(self, release_id: str, environment_id: int,
                           name: str) -> bool:
        """
        Retrieves the current release status of a particular release and environment.

        Arguments:
            release_id (int): ID of the release in Azure DevOps.
            environment_id (int): ID of the environment in a specific release.
            name (str): Name of the release.

        Returns:
            Any of the following results from the Azure DevOps Python SDK:

            canceled: Environment is in canceled state.
            inProgress: Environment is in progress state.
            notStarted: Environment is in not started state.
            partiallySucceeded: Environment is in partially succeeded state.
            queued: Environment is in queued state.
            rejected: Environment is in rejected state.
            scheduled: Environment is in scheduled state.
            succeeded: Environment is in succeeded state.
            undefined: Environment status not set.

        """
        logging.info(f"Project: {self.access_cfg.project}")
        logging.info(f"Release: {release_id}")
        logging.info(f"Environment: {environment_id}")

        release_environment = self.release_client.get_release_environment(
            self.access_cfg.project,
            release_id=release_id,
            environment_id=environment_id)

        return release_environment.status

    def get_latest_successful_release(
            self, name: str, from_env: str,
            target_env: str) -> Union[Tuple[None, None], Tuple[int, int]]:
        """
        Gets the release and environment id for a target environment based off a base environments most
        recent successful or partially successful release.
        Gets a list of all the releases for a specific release pipeline, loops
        through them and for each release looks for a matching environment and status.

        Arguments:
            name: (str) Name of the release pipeline.
            from_env (str): Name of the environment you want the target environment to match.
            target_env (str): Name of the environment that you want the release to run on.

        Returns:
            The ID of the release and environment that was either succeeded or partially succeeded.
            If nothing is found then None, None is returned

        """

        releases = self.release_client.get_releases(self.access_cfg.project,
                                                    search_text=name,
                                                    top=200).value

        for release in releases:

            if release.release_definition.name == name:

                result = self.release_client.get_release(
                    self.access_cfg.project, release_id=release.id)

                target_env_id = self._get_target_environment_id(
                    result.environments, target_env)

                for environment in result.environments:

                    if environment.name == from_env and (
                            environment.status == "succeeded"
                            or environment.status == "partiallySucceeded"
                            or environment.status == "queued"
                            or environment.status == "inProgress"):
                        logging.info(
                            f"Found environment for {name} with id: {environment.id} "
                        )

                        return release.id, target_env_id

        return None, None

    def _get_target_environment_id(self, environments,
                                   environment_name) -> Union[None, int]:
        """
        Internal function used to retrieve the environment id of the target
        environment we want to run a release on.

        Checks to see if InProgress in case the command is run several times by
        mistake.

        environments (List[ReleaseEnvironment]): List of environments to search through.
        environment_name (str): The name of the environment to search for.

        Returns:
            The id of the environment if found. If not found then None is returned.
        """

        for environment in environments:

            if environment.name == environment_name and "inProgress" not in environment.status:
                return environment.id

        return None

    def run_release(self, release_id: int, environment_id: int, release_name: str) -> None:
        """
        Runs the latest succeeded or partially succeeded pipeline associated
        with the object.

        Arguments:
            release_id (int): Id of the release in Azure DevOps.
            environment_id (int): Id of the environment associated to a specific release.
            release_name (str): Name of the release for better logging.

        """
        release_environment = self.release_client.get_release_environment(
            self.access_cfg.project,
            release_id=release_id,
            environment_id=environment_id)

        if release_environment.status != "inProgress" and release_environment.status != "queued":
            start_values = {
                "comment": "Run by Victoria Rebuilder",
                "status": "inProgress"
            }
            self.release_client.update_release_environment(start_values,
                                                           self.access_cfg.project,
                                                           release_id,
                                                           environment_id)
            logging.info(f"Running release {release_id} for {release_name}.")
        else:
            logging.info(
                f"Release {release_id} for {release_name} is currently {release_environment.status} so not starting for this run.")

    def _connect(self):
        """
        Logins to the Azure DevOps API and sets the release_client.

        """
        credentials = BasicAuthentication("", self.access_cfg.access_token)
        connection = Connection(
            base_url=f"https://dev.azure.com/{self.access_cfg.organisation}/",
            creds=credentials)

        self.release_client = connection.clients_v5_1.get_release_client()



"""
Runs and holds the states of deployments for a particular environment.

Parameters:
    from_environment (str): The environment to rebuild from.
    to_environment (str): The environment to rebuild.
    access_cfg (AccessConfig): The configuration to access AzureDevOps.
    deployments (DeploymentConfig): The configuration to process the deployments.
    fresh (bool): If the rebuilder should use the previous state file.

"""
import copy
import logging
import os
import pickle
import sys
import time
from typing import List

from victoria_rebuilder.client import DevOpsClient
from victoria_rebuilder.config import AccessConfig, DeploymentConfig, ReleaseConfig

# The location of the state file.
STATE_FILE = "rebuild"


class Rebuild:
    def __init__(self, from_environment: str, target_environment: str,
                 access_cfg: AccessConfig, deployments: List[DeploymentConfig],
                 resume: bool, auto_retry: bool = False):

        self.deployments = deployments
        self.from_environment = from_environment
        self.target_environment = target_environment
        self.access_cfg = access_cfg
        self.deployments = deployments
        self.auto_retry = auto_retry
        self._load(resume)
        self.client = DevOpsClient(access_cfg)

    def run_deployments(self):
        """
        Rebuilds the environment by running group of deployments.
        Once a deployment has been completed its state is saved and
        the next deployment is run.
        """
        for deployment in self.deployments:

            if not deployment.complete:
                logging.info(f"Running deployment {deployment.stage}")

                deployment.releases = self.run_releases(
                    deployment.releases, self.from_environment,
                    self.target_environment)
                deployment.releases = self.wait_to_complete(
                    deployment.releases, 10)
                deployment.complete = True
                self._save()

            logging.info(f"Deployment {deployment.stage} has completed.")
        self._clean_up()

    def run_releases(self, releases: List[ReleaseConfig], from_environment: str,
                     target_environment: str) -> List[ReleaseConfig]:
        """
        Runs a list of releases associated to a specific deployment and environment.
        If it can't find the release it assumes its not available for that environment
        so it is removed from the list.

        Arguments:
            releases (List[ReleaseConfig]): List of releases that need running.
            from_environment (str): The environment you want to base the target environment on.
            target_environment (str): The environment to run the release on.

        Returns:
            A list of Releases (ReleaseConfig). Releases that weren't found would of been removed.
        """
        for release in releases[:]:
            if not release.complete:
                if not release.release_id:
                    release.release_id, release.environment_id = self.client.get_latest_successful_release(
                        release.name, from_environment, target_environment)

                if release.release_id and release.environment_id:
                    self.client.run_release(release.release_id, release.environment_id, release.name)
                else:
                    logging.info(
                        f"Unable to run release for {release.name}. Either no environment for release or it is currently running."
                    )
                    releases.remove(release)

            self._save()

        return releases

    def wait_to_complete(self, releases: List[ReleaseConfig],
                         interval: int) -> List[ReleaseConfig]:
        """
        Waits for the releases to complete.

        Arguments:
            releases (list[ReleaseConfig]): The list of releases to wait for.
            interval (int): In seconds how often to check if the release is complete.

        Returns:
            A list of releases (ReleaseConfig)
            Once all the releases are complete it returns the list of releases.

        """

        running = True

        while running:

            running = False
            for release in releases:
                time.sleep(interval)

                if not release.complete:
                    release_status = self.client.get_release_status(
                        release.release_id, release.environment_id,
                        release.name)
                    if release_status == "succeeded" or release_status == "partiallySucceeded":
                        logging.info(f"{release.name} is complete.")
                        release.complete = True
                        self._save()
                    elif release_status == "rejected" or release_status == "cancelled":
                        re_run_release = self._re_run_failed_release(release.release_id, release.environment_id,
                                                                     release.name)
                        release.complete = False if re_run_release else True
                    else:
                        running = True

        return releases

    def _re_run_failed_release(self, release_id: int, release_env_id: int, release_name: str) -> bool:
        """
        Asks the user if they want to re run a failed release.

        Arguments:
            release_id (int): The release pipeline ID in Azure DevOpS.
            release_env_id (int): The particular stage ID in Azure DevOps.
            release_name (str): The name of the release pipeline.

        Returns:
            If the user decided to run the pipeline again.
        """
        logging.info(f"{release_name} has not been successful.")

        run_again = self.auto_retry or self._query_yes_no(f"Would you like to run {release_name} again?")

        if run_again:
            self.client.run_release(release_id, release_env_id, release_name)
        else:
            logging.info(f"{release_name} will not be run again. Continuing with the deployment.")

        return run_again

    def _query_yes_no(self, question: str, default="yes") -> bool:
        """
        Ask a yes/no question via input() and return their answer.

        Arguments:
            question (str): A string that is presented to the user.
            default (str): The presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

        Returns:
            The "answer" return value is True for "yes" or False for "no".
        """
        valid = {"yes": True, "y": True, "ye": True,
                 "no": False, "n": False}
        if default is None:
            prompt = " [y/n] "
        elif default == "yes":
            prompt = " [Y/n] "
        elif default == "no":
            prompt = " [y/N] "
        else:
            raise ValueError("invalid default answer: '%s'" % default)

        while True:
            sys.stdout.write(question + prompt)
            choice = input().lower()
            if default is not None and choice == '':
                return valid[default]
            elif choice in valid:
                return valid[choice]
            else:
                sys.stdout.write("Please respond with 'yes' or 'no' "
                                 "(or 'y' or 'n').\n")

    def _load(self, resume: bool) -> None:
        """
        Loads the pickled file of the current object and de-serializes so it can resume
        if there's been a crash or an issue with the pipeline.

        resume (bool): If the rebuilder should use the previous state file.
        """
        if resume:
            if os.path.exists(STATE_FILE):
                with open(STATE_FILE, 'rb') as rebuild_obj_file:
                    # nosec
                    loaded_dict = pickle.load(rebuild_obj_file)
                    self.__dict__.update(loaded_dict)
            else:
                logging.info("Unable to find rebuild file. Assuming fresh run. ")
        else:
            self._clean_up()
            logging.info("Fresh run so have removed the previous state file.")

    def _save(self) -> None:
        """
        Creates a deep copy of the current state of the object, removes the client
        connection so it can be pickled, pickles self and saves it to a file.
        """
        with open(STATE_FILE, 'wb') as rebuild_obj_file:
            current_state = copy.copy(self.__dict__)

            current_state['client'] = None

            pickle.dump(current_state, rebuild_obj_file)

    def _clean_up(self) -> None:
        """
        If the deployment has been successful then the state file
        is removed.
        """
        if os.path.exists(STATE_FILE):
            os.remove(STATE_FILE)



from victoria.plugin import Plugin
from .config import RebuilderSchema
from . import cli

# this object is loaded by Victoria and used as the plugin entry point
plugin = Plugin(name="rebuilder",
                cli=cli.rebuilder,
                config_schema=RebuilderSchema())



"""cli.py

This is the module that contains the Click CLI for V.I.C.T.O.R.I.A rebuilder

Author:
    Alex Potter-Dixon <apotter-dixon@glasswallsolutions.com>
"""

import logging

import click

from .config import RebuilderConfig, AccessConfig
from .rebuild import Rebuild


@click.group()
@click.pass_obj
def rebuilder(cfg: RebuilderConfig):
    """
    The Rebuilder allows the rebuilding of environments via CLI.
    """
    pass


@rebuilder.command()
@click.argument('from_env', nargs=1, type=str)
@click.argument('to_env', nargs=1, type=str)
@click.option(
    '-r',
    '--resume',
    is_flag=True,
    help="If you want the rebuilder to use the previous state file.")
@click.pass_obj
def copy(cfg: RebuilderConfig, from_env: str, to_env: str,
         resume: bool) -> None:
    """
    CLI call for rebuilding an environment based off another environment.
    Arguments:
        cfg (RebuilderConfig): The rebuilder config.
        from_env (str): The environment to rebuild from in Azure DevOps.
        to_env (str): The environment to rebuild.
        resume (bool): If the rebuilder should resume using previous state file.

    """
    logging.info(
        f"Rebuilding environments {from_env} from environment: {to_env}")

    logging.info(f"Rebuilding environment {to_env}.")

    access_cfg = retrieve_access_config(cfg)
    env_rebuild = Rebuild(from_env.lower(), to_env.lower(), access_cfg,
                          cfg.deployments, resume)

    env_rebuild.run_deployments()

    logging.info(f"Finished running deployments to {to_env}.")


@rebuilder.command()
@click.argument('env', nargs=1, type=str)
@click.option(
    '-r',
    '--resume',
    is_flag=True,
    help="If you want the rebuilder to use the previous state file.")
@click.option(
    '-a',
    '--auto-retry',
    is_flag=True,
    help="If a release fails to deploy, instead of prompting the user for a y/n on retry, it automatically retries "
         "the deployment.")
@click.pass_obj
def rebuild(cfg: RebuilderConfig, env: str, resume: bool, auto_retry: bool) -> None:
    """
    CLI call for rebuilding a specific kubernetes environment
    Arguments:
        cfg (RebuilderConfig): The rebuilder config.
        env (str): Environment to rebuild.
        resume (bool): If the rebuilder should use the previous state file.
        auto_retry (bool): If a release fails to deploy, instead of prompting the user for a y/n on retry, it automatically retries the deployment.
    """
    logging.info(f"Rebuilding environment {env}.")

    access_cfg = retrieve_access_config(cfg)
    env_rebuild = Rebuild(env.lower(), env.lower(), access_cfg,
                          cfg.deployments, resume, auto_retry)

    env_rebuild.run_deployments()

    logging.info(f"Finished running deployments to {env}.")


def retrieve_access_config(cfg: RebuilderConfig) -> AccessConfig:
    try:
        encryption_provider = cfg.victoria_config.get_encryption()
    except AttributeError:
        logging.error(
            "Please specify 'provider' in the Victoria settings.")
        raise SystemExit(1)
    else:
        access_token = encryption_provider.decrypt_str(cfg.access.access_token)
        email = encryption_provider.decrypt_str(cfg.access.email)
        organisation = encryption_provider.decrypt_str(cfg.access.organisation)
        project = encryption_provider.decrypt_str(cfg.access.project)

        if None in [access_token, email, organisation, project]:
            logging.error(
                "Invalid 'access' settings provided. Unable to decrypt.")
            raise SystemExit(1)

        return AccessConfig(access_token=access_token, email=email, organisation=organisation, project=project)



"""setup.py

Used for creation and DESTRUCTION of SAAS environments

Author:
    Alex Potter-Dixon <apotter-dixon@glasswallsolutions.com>
"""
from setuptools import setup, find_packages


def repo_file_as_string(file_path: str) -> str:
    with open(file_path, "r") as repo_file:
        return repo_file.read()


setup(
    dependency_links=[],
    install_requires=[
        "click", "marshmallow", "pyyaml", "requests", "azure-devops",
        "victoria"
    ],
    name="victoria_rebuilder",
    version="0.2.1",
    description=
    "Victoria Plugin that allows the creation and DESTRUCTION of SAAS on the cloud.",
    long_description=repo_file_as_string("README.md"),
    long_description_content_type="text/markdown",
    author="Alex Potter-Dixon",
    packages=find_packages(),
)